//
//  OSDocumentManager.m
//  Common
//
//  Created by Duncan Groenewald on 15/07/13.
//  Copyright (c) 2013 Duncan Groenewald. All rights reserved.
//


#import "OSxCDStackManager.h"


NSString* const ICloudStateUpdatedNotification = @"ICloudStateUpdatedNotification";

NSString* const OSFileDeletedNotification = @"OSFileDeletedNotification";
NSString* const OSFileCreatedNotification = @"OSFileCreatedNotification";
NSString* const OSFileClosedNotification = @"OSFileClosedNotification";
NSString* const OSFilesUpdatedNotification = @"OSFilesUpdatedNotification";
NSString* const OSDataUpdatedNotification = @"OSCoreDataUpdated";
NSString* const OSStoreChangeNotification = @"OSCoreDataStoreChanged";
NSString* const OSStoreReadyNotification = @"OSCoreDataStoreReady";
NSString* const OSStoreUpgradingNotification = @"OSCoreDataStoreUpgrading";
NSString* const OSStoreUpgradeStatusNotification = @"OSCoreDataStoreUpgradeStatus";
NSString* const OSBackgroundTaskDoneNotification = @"OSBackgroundTaskDone";

// The name of the file that contains the store identifier.
static NSString *DocumentMetadataFileName = @"DocumentMetadata.plist";

// The name of the file package subdirectory that contains the Core Data store when local.
static NSString *StoreDirectoryComponentLocal = @"StoreContent";

// The name of the file package subdirectory that contains the Core Data store when in the cloud. The Core Data store itself should not be synced directly, so it is placed in a .nosync directory.
static NSString *StoreDirectoryComponentCloud = @"StoreContent.nosync";

static NSString *StoreFileName = @"persistentStore";


static OSxCDStackManager* __sharedManager;

// Just a class to store details of a file.
@implementation FileRepresentation

- (id)initWithFileName:(NSString *)fileName url:(NSURL *)url
{
    self = [super init];
    if (self) {
        _fileName = fileName;
        _url = url;
        _fileDate = @"";
        _percentDownloaded = nil;
    }
    
    return self;
}
- (id)initWithFileName:(NSString *)fileName url:(NSURL *)url date:(NSString*)fileDate
{
    self = [super init];
    if (self) {
        _fileName = fileName;
        _url = url;
        _fileDate = fileDate;
        _percentDownloaded = nil;
    }
    
    return self;
}
- (id)initWithFileName:(NSString *)fileName url:(NSURL *)url percentDownloaded:(NSNumber*)percent
{
    self = [super init];
    if (self) {
        _fileName = fileName;
        _url = url;
        _fileDate = @"";
        _percentDownloaded = percent;
        _fileDate = [self modifiedDate];
    }
    
    return self;
}
- (BOOL)isEqual:(FileRepresentation*)object
{
    return [object isKindOfClass:[FileRepresentation class]] && [_fileName isEqual:object.fileName];
}
- (NSString*)modifiedDate {
    NSError *er = nil;
    NSDate *date = nil;
    if ([_url getResourceValue:&date forKey:NSURLContentModificationDateKey error:&er] == YES)
    {
        NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
        [dateFormatter setDateStyle:NSDateFormatterMediumStyle];
        [dateFormatter setTimeStyle:NSDateFormatterMediumStyle];
        return [dateFormatter stringFromDate:date];
    } else {
        return @"" ;
    }
}
@end



@implementation OSxCDStackManager
{
    BOOL _isCloudEnabled;
    NSURL* _dataDirectoryURL;
}

@synthesize isCloudEnabled = _isCloudEnabled;
@synthesize dataDirectoryURL = _dataDirectoryURL;
@synthesize ubiquityID = _ubiquityID;

+ (OSxCDStackManager*)sharedManager
{
    if (!__sharedManager) {
        __sharedManager = [[OSxCDStackManager alloc] init];
    }
    
    return __sharedManager;
}

- (id)init
{
    if ((self = [super init])) {
        
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(checkUserICloudPreferenceAndSetupIfNecessary) name:NSUbiquityIdentityDidChangeNotification object:nil];
        
        // We store everything locally
        [self updateFileStorageContainerURL];
        
        _persistentStoreCoordinator = nil;
        _useICloudStorage = -1;
        _import_or_save = NO;
        _load_seed_data = NO;
        _use_old_model = NO;
        _isUpgrading = NO;
        _startUpgrade = NO;
        // Used during development to delete all files off a device
        //if (_clearFiles)
        //    [self clearAllFiles];
    }
    _isBusy = NO;
    return self;
}
+ (NSString *)persistentStoreName {
    return @"persistentStore";
}
+ (NSString *)icloudPersistentStoreName {
    return @"persistentStore_ICLOUD";
}
- (NSURL *)icloudStoreURL {
    return [[self applicationFilesDirectory] URLByAppendingPathComponent:[OSxCDStackManager icloudPersistentStoreName]];
}
- (NSURL *)backupStoreURL {
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    [dateFormatter setDateFormat:@"yyyyMMddHHmmss"];
    NSString *dateString = [dateFormatter stringFromDate:[NSDate date]];
    
    
    NSString *fileName = [NSString stringWithFormat:@"%@_Backup_%@",[OSxCDStackManager persistentStoreName], dateString];
    
    return [[self applicationFilesDirectory] URLByAppendingPathComponent:fileName];
}
- (NSURL *)localUbiquitySupportURL {
    return [[self applicationFilesDirectory] URLByAppendingPathComponent:@"CoreDataUbiquitySupport"];
}
/*! Returns the CoreData directory in the ubiquity container
 
 @returns The URL for the CoreData directory in ubiquity container
 */
- (NSURL *)icloudContainerURL {
    
    NSURL *iCloudURL = [[NSFileManager defaultManager] URLForUbiquityContainerIdentifier:_ubiquityID];
    return [[iCloudURL URLByAppendingPathComponent:@"CoreData"] URLByAppendingPathComponent:[OSxCDStackManager icloudPersistentStoreName]];
}
- (NSURL *)localStoreURL {
    return [[self applicationFilesDirectory] URLByAppendingPathComponent:[OSxCDStackManager persistentStoreName]];
}
- (void)setUbiquityID:(NSString*)ubiquityID
{
    _ubiquityID = ubiquityID;
}
/*! Posts a notification that files have been changed
 */
- (void)postFileUpdateNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSFilesUpdatedNotification
                                                            object:self];
    });
}
/*! Posts a notification that file has been deleted
 */
- (void)postFileDeletedNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSFileDeletedNotification
                                                            object:self];
    });
}
/*! Posts a notification that file has been deleted
 */
- (void)postFileCreatedNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSFileCreatedNotification
                                                            object:self];
    });
}
/*! Posts a notification that file has been deleted
 */
- (void)postFileClosedNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSFileClosedNotification
                                                            object:self];
    });
}
/*! Posts a notification that background job has been completed */
- (void)postJobDoneNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSBackgroundTaskDoneNotification
                                                            object:self];
    });
}
/*! Posts a notification that a store upgrade is being performed */
- (void)postStoreUpgradingNotification {
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSStoreUpgradingNotification
                                                            object:self];
    });
}
/*! Posts a notification about a store upgrade status event */
- (void)postStoreUpgradeStatusNotification:(NSString*)message {
    
    dispatch_async(dispatch_get_main_queue(), ^{
        [[NSNotificationCenter defaultCenter] postNotificationName:OSStoreUpgradeStatusNotification
                                                            object:self userInfo:[[NSDictionary alloc] initWithObjectsAndKeys:message, @"message", nil]];
    });
}
/*!  This method is called when the user selects whether they want to delete any files from the device in response to disabling iCloud
 for the app
 
 @param deleteICloudFiles A bool value YES or NO.  Used in combination with setIsCloudEnabled to take actions to copy files to/from iCloud
 */
- (void)setDeleteICloudFiles:(BOOL)deleteICloudFiles
{
    _deleteICloudFiles = deleteICloudFiles;
}
/*! This gets called to enable or disable use of iCloud.  This is only called when the App specific setting is changed
 because if its the global setting we don't do anything anyway.  If the global iCloud setting is turned OFF then the app
 would already have lost access to iCloud so we can't migrate files to local storage.
 If the global iCloud setting is set to ON then we need to check if the local setting is set to ON or OFF before we do anything.
 
 When iCloud is enabled then this function triggers the sharing of any local files to iCloud by migrating the files using [PSC migratePersistentStore]
 At the same time this function will trigger the metadata scan for iCloud files so that any files in iCloud appear locally.  For these iCloud files to
 appear locally we have to build local stores from the iCloud transaction logs using standard [PSC addPersistentStore].
 
 When iCloud option is disabled we need to migrate the iCloud stores to local stores and then remove the stores from iCloud.
 
 @param  isCloudEnabled A bool value YES or NO indicating whether iCloud is enabled or disabled.
 */
- (void)setIsCloudEnabled:(BOOL)isCloudEnabled
{
    FLOG(@"setIsCloudEnabled:%@ called", (isCloudEnabled ? @"YES" : @"NO"));
    
    _isCloudEnabled = isCloudEnabled;
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
        
        [self migrateFilesIfRequired];
        
        if ([self checkCDModelVersion]) {
            // We must set a flag so the old version of the model gets used to migrate the existing store to
            // local
            [self showBackgroundTaskActive];
            [self postStoreUpgradingNotification];
            
            // Now wait until user has signalled START
            
            while (!_startUpgrade) {
                [self postStoreUpgradeStatusNotification:@"Waiting to start..."];
                sleep(2);
            }
            
            _isUpgrading = YES;
            _use_old_model = YES;
            _managedObjectModel = nil;
            //Store needs to be upgraded so migrate to local before opening
            [self postStoreUpgradeStatusNotification:@"Migrating iCloud store to local..."];
            if ([self migrateICloudFileToLocal]) {
                [self postStoreUpgradeStatusNotification:@"Removing iCloud store..."];
                _use_old_model = NO;
                _managedObjectModel = nil;
                self.storeURL = [self localStoreURL];
                [self postStoreUpgradeStatusNotification:@"Upgrading local store..."];
                [self openPersistentStore];
                _managedObjectContext = nil;
                _persistentStoreCoordinator = nil;
                [self postStoreUpgradeStatusNotification:@"Migrating upgraded store back to iCloud..."];
                if ([self migrateLocalFileToICloud]) {
                    self.storeURL = [self icloudStoreURL];
                } else {
                    FLOG(@" error migrating store back to iCloud after upgrade");
                    [self postStoreUpgradeStatusNotification:@"Error migrating store back to iCloud."];
                }
            } else {
                [self postStoreUpgradeStatusNotification:@"Error migrating iCloud store to local."];
                FLOG(@" error migrating store to local for migration");
                abort();
            }
            _isUpgrading = NO;
            [self showBackgroundTaskActive];
            [self postStoreUpgradeStatusNotification:@"Upgrade complete."];
            [self postStoreUpgradeStatusNotification:@"Opening upgraded iCloud store."];

        }
        
        [self openPersistentStore];
        
    });
    
}
/*! Migrates any files based on the current settings.
 
 */
- (void)migrateFilesIfRequired {
    FLOG(@"migrateFilesIfRequired called...");
    // Setting has changed to take the appropriate action
    
    if (_isCloudEnabled) {
        // iCloud has been enabled so migrate local files to iCloud
        FLOG(@" iCloud has been enabled so migrate local files to iCloud if they exist");
        
        if ([self localStoreExists]) {
            FLOG(@" Local store exists so migrate it");
            
            if ([self migrateLocalFileToICloud]) {
                FLOG(@" Local store migrated to iCloud successfully");
                self.storeURL = [self icloudStoreURL];
                
            } else {
                FLOG(@" Local store migration to iCloud FAILED because iCloud store already there");
                
                // Do nothing because we have posted an alert asking the user what to do and we will respond to that
                self.storeURL = nil;
            }
            
        } else {
            FLOG(@" No local store exists");
            self.storeURL = [self icloudStoreURL];
        }
        
    } else {
        // iCloud has been disabled so check whether to keep or delete them
        FLOG(@" iCloud has been disabled so check if there are any iCloud files and delete or migrate them");
        
        if ([self iCloudStoreExists]) {
            FLOG(@" iCloud store exists");
            
            if (_deleteICloudFiles) {
                FLOG(@" delete iCloud Files");
                // DG Need to add Code for this !
                [self removeICloudStore];
                [self deregisterForStoreChanges];
                _persistentStoreCoordinator = nil;
                _managedObjectContext = nil;
                self.storeURL = [self localStoreURL];
            } else {
                FLOG(@" migrate iCloud Files");
                if ([self migrateICloudFileToLocal]) {
                    FLOG(@" iCloud store migrated to Local successfully");
                    self.storeURL = [self localStoreURL];
                } else {
                    FLOG(@" iCloud store migration to Local FAILED");
                    self.storeURL = [self localStoreURL];
                }
            }
        } else {
            FLOG(@" no iCloud store exists so no migration required");
            self.storeURL = [self localStoreURL];
        }
    }
    
}
/*! Migrates and iCloud file to a Local file by creating a OSManagedDocument
 and calling the moveDocumentToLocal method.  The document knows how to move itself.
 
 @param fileURL The URL of the file to be moved
 */
- (BOOL)migrateICloudFileToLocal {
    FLOG(@" migrateICloudFileToLocal");
    
    
    // Now check if the file is in iCloud
    if (![self localStoreExists]) {
        
        return [self moveStoreToLocal];
        
    } else {
        
        FLOG(@" error migrateICloudFileToLocal because Local file already exists!");
        return NO;
        
    }
}

/*! Migrates a Local file to iCloud file by creating a OSManagedDocument
 and calling the moveDocumentToICloud method.  The document knows how to move itself.
 
 @param fileURL The URL of the file to be moved
 */
- (bool)migrateLocalFileToICloud {
    FLOG(@" migrateLocalFileToiCloud");
    
    // Now check if the file is already in iCloud
    if (![self iCloudStoreExists]) {
        
        return [self moveStoreToICloud];
        
    } else {
        
        FLOG(@" Warning migrating local file to iCloud but iCloud file already exists, choose an option!");
        
        FLOG(@" _cloudMergeChoiceAlert being processed");
        
        if ([self mergeWithICloudChoiceAlert]) {
            FLOG(@" user selected to merge with iCloud file");
            [self deregisterForStoreChanges];
            _persistentStoreCoordinator = nil;
            _managedObjectContext = nil;
            [self moveStoreToICloud];
            [self postStoreChangedNotification];
        }
        else {
            FLOG(@" user selected to use iCloud file");
            [self deregisterForStoreChanges];
            _persistentStoreCoordinator = nil;
            _managedObjectContext = nil;
            self.storeURL = [self icloudStoreURL];
            
            // Always make a backup of the local store before migrating to iCloud
            [self backupLocalStore];
            
            [self deleteLocalStore];
            [self postStoreChangedNotification];
        }
        
        
        return YES;
        
    }
}

/*! Checks for the existence of any documents with _UUID_ in the filename.  These documents are documents shared in
 iCloud.  Returns YES if any are found.
 
 @return Returns YES of any documents are found or NO if none are found.
 */
- (bool)iCloudStoreExists {
    FLOG(@" called");
    
    FLOG(@"  icloudStoreURL is %@", [[self icloudContainerURL] path]);
    
    BOOL isDir;
    BOOL fileExists = [[NSFileManager defaultManager] fileExistsAtPath:[[self icloudContainerURL] path] isDirectory:&isDir];
    
    FLOG(@"  icloudContainerURL %@", (fileExists ? @"exists" : @"does not exist"));
    
    return fileExists;
}
/*! Checks for the existence of any documents without _UUID_ in the filename.  These documents are local documents only.  Returns YES if any are found.
 
 @return Returns YES of any documents are found or NO if none are found.
 */
- (bool)localStoreExists {
    FLOG(@"localStoreExists called");
    
    FLOG(@"  localStoreURL is %@", [[self localStoreURL] path]);
    BOOL isDir;
    BOOL fileExists = [[NSFileManager defaultManager] fileExistsAtPath:[[self localStoreURL] path] isDirectory:&isDir];
    
    FLOG(@"  localStoreURL %@", (fileExists ? @"exists" : @"does not exist"));
    
    return fileExists;
}

/*! Returns the CoreData directory in the ubiquity container
 
 @returns The URL for the CoreData directory in ubiquity container
 */
- (NSURL*)iCloudCoreDataURL {
    NSURL *iCloudURL = [[NSFileManager defaultManager] URLForUbiquityContainerIdentifier:_ubiquityID];
    return [iCloudURL URLByAppendingPathComponent:@"CoreData"];
}

/*! Returns the /Documents directory for the app
 
 @returns Returns the URL for the apps /Documents directory
 */
- (NSURL*)documentsDirectoryURL
{
    _dataDirectoryURL = [NSURL fileURLWithPath:NSHomeDirectory() isDirectory:YES];
    return [_dataDirectoryURL URLByAppendingPathComponent:@"Documents"];
}

- (void)updateFileStorageContainerURL
{
    // Perform the asynchronous update of the data directory and document directory URLs
    _dataDirectoryURL = [NSURL fileURLWithPath:NSHomeDirectory() isDirectory:YES];
}

/*! Called to determine whether the user preference "Use iCloud" has been set to YES and iCloud
 is available (user logged in and Documents & Data ON).  If this returns YES then create new files
 with UUID and iCloud options.  If this returns NO create new files without UUID or iCloud options.
 
 @return Returns YES if both app iCloud "Use iCloud" preferences is YES and iCloud is available.
 */
- (bool)canUseICloud {
    
    // Get the user preference setting
    NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
    //NSString* userICloudChoiceSet = [userDefaults stringForKey:self.cloudPreferenceSet];
    
    bool userICloudChoice = [userDefaults boolForKey:self.cloudPreferenceKey];
    FLOG(@" User preference for %@ is %@", self.cloudPreferenceSet, (userICloudChoice ? @"YES" : @"NO"));
    
    if ([[NSFileManager defaultManager] ubiquityIdentityToken] && userICloudChoice)
        return YES;
    else
        return NO;
    
}
/*! Called to determine whether iCloud is available (user logged in and Documents & Data ON).
 
 @return Returns YES if iCloud account is active and Data & Documents is ON
 */
- (bool)isICloudAvailable {
    
    
    if ([[NSFileManager defaultManager] ubiquityIdentityToken])
        return YES;
    else
        return NO;
    
}
- (bool)isBusy {
    return _isBusy;
}

// We only care if the one we have open is changing
- (void)registerForStoreChanges:(NSPersistentStoreCoordinator*)storeCoordinator {
    //LOG(@"registerForStoreChanges called");
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesWillChange:) name:NSPersistentStoreCoordinatorStoresWillChangeNotification object:storeCoordinator];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesDidChange:) name:NSPersistentStoreCoordinatorStoresDidChangeNotification object:storeCoordinator];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesDidImport:) name:NSPersistentStoreDidImportUbiquitousContentChangesNotification object:storeCoordinator];
    
    
}
- (void)deregisterForStoreChanges {
    //LOG(@"degisterForStoreChanges called");
    [[NSNotificationCenter defaultCenter] removeObserver:self  name:NSPersistentStoreCoordinatorStoresWillChangeNotification object:nil];
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSPersistentStoreCoordinatorStoresDidChangeNotification object:nil];
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSPersistentStoreDidImportUbiquitousContentChangesNotification object:nil];
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSManagedObjectContextDidSaveNotification object:nil];
    
}
// NB - this may be called from a background thread so make sure we run on the main thread !!
// This is when store files are being switched from fallback to iCloud store
- (void)storesWillChange:(NSNotification*)n {
    FLOG(@"storesWillChange called - >>>>>>>>>>>>>>>>>>>>>>>>>>>>");
    
    // Check type of transition
    NSNumber *type = [n.userInfo objectForKey:NSPersistentStoreUbiquitousTransitionTypeKey];
    
    FLOG(@" transition type is %@", type);
    
    if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeInitialImportCompleted) {
        
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeInitialImportCompleted");
        
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeAccountAdded) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeAccountAdded");
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeAccountRemoved) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeAccountRemoved");
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeContentRemoved) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeContentRemoved");
    }
    
    
    [[NSOperationQueue mainQueue] addOperationWithBlock:^ {
        
        NSArray *addedStores = [n.userInfo objectForKey:NSAddedPersistentStoresKey];
        NSArray *removedStores = [n.userInfo objectForKey:NSRemovedPersistentStoresKey];
        NSArray *changedStores = [n.userInfo objectForKey:NSUUIDChangedPersistentStoresKey];
        
        FLOG(@" added stores are %@", addedStores);
        FLOG(@" removed stores are %@", removedStores);
        FLOG(@" changed stores are %@", changedStores);
        
        NSError *error;
        if ([self.managedObjectContext hasChanges]) {
            [self.managedObjectContext save:&error];
        }
        
        [self.managedObjectContext reset];
        // Reset user Interface - i.e. lock the user out!
        _storesChanging = YES;
        
        //DG Turn this off for now
        //[self showStoresChangingAlert];
    }];
    
}
// NB - this may be called from a background thread so make sure we run on the main thread !!
// This is when store files are being switched from fallback to iCloud store
- (void)storesDidChange:(NSNotification*)n {
    FLOG(@"storesDidChange called - >>>>>>>>>>>>>>>>>>>>>>>>>>>>");
    
    // Check type of transition
    NSNumber *type = [n.userInfo objectForKey:NSPersistentStoreUbiquitousTransitionTypeKey];
    
    FLOG(@" userInfo is %@", n.userInfo);
    FLOG(@" transition type is %@", type);
    
    if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeInitialImportCompleted) {
        
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeInitialImportCompleted");
        
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeAccountAdded) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeAccountAdded");
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeAccountRemoved) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeAccountRemoved");
    } else if (type.intValue == NSPersistentStoreUbiquitousTransitionTypeContentRemoved) {
        FLOG(@" transition type is NSPersistentStoreUbiquitousTransitionTypeContentRemoved");
    }
    
    [[NSOperationQueue mainQueue] addOperationWithBlock:^ {
        
        _storesChanging = NO;
        
        if (type != nil && type.intValue == NSPersistentStoreUbiquitousTransitionTypeContentRemoved) {
            
            // [self showStoreRemovedAlert];
            //_persistentStoreCoordinator = nil;
            _managedObjectContext = nil;
            
            FLOG(@" iCloud store was removed! Wait for empty store");
        }
        [self postStoreChangedNotification];
        // Refresh user Interface
        [self createTimer];
        
    }];
}
/*
 - (void)showStoreRemovedAlert {
 
 if (_storesUpdatingAlert == nil) {
 NSString * message = [NSString stringWithFormat:@"Please wait for a new store to be created."];
 _storesUpdatingAlert = [[UIAlertView alloc] initWithTitle:@"Store file has been removed" message:message delegate:nil cancelButtonTitle:nil otherButtonTitles:nil];
 FLOG(@" XXX showStoreRemovedAlert %@", message);
 [_storesUpdatingAlert show];
 }
 
 }
 */
// NB - this may be called from a background thread so make sure we run on the main thread !!
// This is when transactoin logs are loaded
- (void)storesDidImport:(NSNotification*)notification {
    
    [[NSOperationQueue mainQueue] addOperationWithBlock:^ {
        FLOG(@"storesDidImport ");
        /* Process new/changed objects here and remove any unwanted items or duplicates prior to merging with current context
         //
         NSSet *updatedObjectIDs = [[notification userInfo] objectForKey:NSUpdatedObjectsKey];
         NSSet *insertedObjectIDs = [[notification userInfo] objectForKey:NSInsertedObjectsKey];
         NSSet *deletedObjectIDs = [[notification userInfo] objectForKey:NSDeletedObjectsKey];
         
         // Iterate over all the new, changed or deleted ManagedObjectIDs and get the NSManagedObject for the corresponding ID:
         // These come from another thread so we can't reference the objects directly
         
         for(NSManagedObjectID *managedObjectID in updatedObjectIDs){
         NSManagedObject *managedObject = [_managedObjectContext objectWithID:managedObjectID];
         }
         
         // Check is some object is equal to an inserted or updated object
         if([myEntity.objectID isEqual:managedObject.objectID]){}
         */
        if (self.managedObjectContext) {
            [self.managedObjectContext mergeChangesFromContextDidSaveNotification:notification];
        }
        _import_or_save = YES;
        
        [self createTimer];
        
        
    }];
}
// NB - this may be called from a background thread so make sure we run on the main thread !!
// This is when transactoin logs are loaded
- (void)storesDidSave:(NSNotification*)notification {
    
    // Ignore any notifications from the main thread because we only need to merge data
    // loaded from other threads.
    if ([NSThread isMainThread]) return;
    
    [[NSOperationQueue mainQueue] addOperationWithBlock:^ {
        FLOG(@"storesDidSave ");
        if (self.managedObjectContext) {
            [self.managedObjectContext mergeChangesFromContextDidSaveNotification:notification];
        }
        _import_or_save = YES;
        [self createTimer];
        
    }];
}

- (void)createTimer {
    if (self.iCloudUpdateTimer == nil) {
        self.iCloudUpdateTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
                                                                  target:self
                                                                selector:@selector(notifyOfCoreDataUpdates)
                                                                userInfo:nil
                                                                 repeats:NO];
    } else {
        [self.iCloudUpdateTimer setFireDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];
    }
}
- (void)notifyOfCoreDataUpdates {
    FLOG(@"notifyOfCoreDataUpdates called");
    //if (_storesUpdatingAlert)
    //    [_storesUpdatingAlert dismissWithClickedButtonIndex:0 animated:YES];
    
    if (_import_or_save) {
        NSError *error;
        //if ([self.managedObjectContext hasChanges]) {
        [self.managedObjectContext save:&error];
        //}
        _import_or_save = NO;
    }
    
    [self.iCloudUpdateTimer invalidate];
    self.iCloudUpdateTimer = nil;
    [self postUIUpdateNotification];
}
- (void)postUIUpdateNotification {
    [[NSNotificationCenter defaultCenter] postNotificationName:OSDataUpdatedNotification
                                                        object:self];
}
- (void)postStoreChangedNotification {
    [[NSNotificationCenter defaultCenter] postNotificationName:OSStoreChangeNotification
                                                        object:self];
}
- (void)postStoreReadyNotification {
    [[NSNotificationCenter defaultCenter] postNotificationName:OSStoreReadyNotification
                                                        object:self];
}

- (void)documentStateChanged:(NSNotification*)notification
{
    /*
     switch ([[notification object] documentState]) {
     case UIDocumentStateNormal:
     //FLOG(@"UIDocumentStateNormal");
     break;
     case UIDocumentStateClosed:
     FLOG(@"UIDocumentStateClosed %@", notification);
     break;
     case UIDocumentStateInConflict:
     FLOG(@"UIDocumentStateInConflict %@", notification);
     break;
     case UIDocumentStateSavingError:
     FLOG(@"UIDocumentStateSavingError %@", notification);
     break;
     case UIDocumentStateEditingDisabled:
     FLOG(@"UIDocumentStateEditingDisabled %@", notification);
     break;
     }
     */
}
- (void)addInitialData {
    FLOG(@"addInitialData called");
}

/*! Checks whether the ubiquity token has changed and if so it means the iCloud login has changed since the application was last
 active.  If the user has signed out then they will loose access to their iCloud documents so tell them to log back in to
 access those documents.
 
 @param currenToken The current ubiquity identity.
 */
- (void)checkUbiquitousTokenFromPreviousLaunch:(id)currentToken
{
    // Fetch a previously stored value for the ubiquity identity token from NSUserDefaults.
    // That value can be compared to the current token to determine if the iCloud login has changed since the last launch of our application
    
    NSData* oldTokenData = [[NSUserDefaults standardUserDefaults] objectForKey:_ubiquityIDToken];
    id oldToken = oldTokenData ? [NSKeyedUnarchiver unarchiveObjectWithData:oldTokenData] : nil;
    if (oldTokenData && ![oldToken isEqual:currentToken]) {
        // If we had a token, we were signed in before.
        // If the token has change, a signout has occurred - either switching to another account or deleting iCloud entirely.
        
        // UIAlertView* alert = [[UIAlertView alloc] initWithTitle:@"iCloud Sign-Out" message:@"You have signed out of the iCloud account previously used to store documents. Sign back in to access those documents" delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
        // [alert show];
    }
}
- (void)showMigratingAlert:(NSString*)str {
    /*
     if (_migratingAlert == nil) {
     NSString * message = [NSString stringWithFormat:@"Please wait while your files get migrated %@ iCloud", str];
     _migratingAlert = [[UIAlertView alloc] initWithTitle:@"File migration in progress" message:message delegate:nil cancelButtonTitle:nil otherButtonTitles:nil];
     FLOG(@" XXX showMigratingAlert %@", message);
     [_migratingAlert show];
     }
     */
}
- (void)dismissMigratingAlert {
    /*
     if (_migratingAlert != nil) {
     FLOG(@" XXX dismissMigratingAlert");
     [_migratingAlert dismissWithClickedButtonIndex:0 animated:YES];
     _migratingAlert = nil;
     }
     */
}
/*
 - (void)alertView:(UIAlertView*)alertView didDismissWithButtonIndex:(NSInteger)buttonIndex
 {
 if (alertView == _cloudChoiceAlert)
 {
 FLOG(@" _cloudChoiceAlert being processed");
 if (buttonIndex == 1) {
 FLOG(@" user selected iCloud files");
 [[NSUserDefaults standardUserDefaults] setBool:YES forKey:_cloudPreferenceKey];
 [[NSUserDefaults standardUserDefaults] setValue:@"YES" forKey:_cloudPreferenceSet];
 _useICloudStorage = YES;
 [[NSUserDefaults standardUserDefaults] synchronize];
 
 [self setIsCloudEnabled:YES];
 //[self postFileUpdateNotification];
 }
 else {
 FLOG(@" user selected local files");
 [[NSUserDefaults standardUserDefaults] setBool:NO forKey:_cloudPreferenceKey];
 [[NSUserDefaults standardUserDefaults] setValue:@"YES" forKey:_cloudPreferenceSet];
 _useICloudStorage = NO;
 [[NSUserDefaults standardUserDefaults] synchronize];
 
 [self setIsCloudEnabled:NO];
 }
 }
 if (alertView == _cloudChangedAlert)
 {   FLOG(@" _cloudChangedAlert being processed");
 if (buttonIndex == 0) {
 FLOG(@" 'Keep using iCloud' selected");
 FLOG(@" turn Use iCloud back ON");
 [[NSUserDefaults standardUserDefaults] setBool:YES forKey:_cloudPreferenceKey];
 [[NSUserDefaults standardUserDefaults] synchronize];
 _useICloudStorage = YES;
 [self setIsCloudEnabled:YES];
 
 }
 else if (buttonIndex == 1) {
 FLOG(@" 'Keep on My iPhone' selected");
 FLOG(@" copy to local storage");
 _useICloudStorage = NO;
 [self setDeleteICloudFiles:NO];
 [self setIsCloudEnabled:NO];
 
 }else if (buttonIndex == 2) {
 FLOG(@" 'Delete from My iPhone' selected");
 FLOG(@" delete copies from iPhone");
 _useICloudStorage = NO;
 [self setDeleteICloudFiles:YES];
 [self setIsCloudEnabled:NO];
 }
 [self postFileUpdateNotification];
 }
 if (alertView == _cloudMergeChoiceAlert)
 {
 FLOG(@" _cloudMergeChoiceAlert being processed");
 if (buttonIndex == 0) {
 FLOG(@" user selected to use iCloud file");
 _persistentStoreCoordinator = nil;
 _managedObjectContext = nil;
 self.storeURL = [self icloudStoreURL];
 [self deleteLocalStore];
 [self postStoreChangedNotification];
 }
 else {
 FLOG(@" user selected to merge with iCloud file");
 _persistentStoreCoordinator = nil;
 _managedObjectContext = nil;
 [self moveStoreToICloud];
 [self postStoreChangedNotification];
 }
 }
 
 }
 */
- (void)storeCurrentUbiquityToken:(id)currentToken
{
    // Write the ubquity identity token to NSUserDefaults if it exists.
    // Otherwise, remove the key.
    
    if (currentToken) {
        NSData* newTokenData = [NSKeyedArchiver archivedDataWithRootObject:currentToken];
        [[NSUserDefaults standardUserDefaults] setObject:newTokenData forKey:_ubiquityIDToken];
    }
    else {
        [[NSUserDefaults standardUserDefaults] removeObjectForKey:_ubiquityIDToken];
    }
}
- (bool)userICloudChoice {
    NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
    [userDefaults synchronize];
    
    NSString* userICloudChoiceSet = [userDefaults stringForKey:_cloudPreferenceSet];
    bool userICloudChoice = [userDefaults boolForKey:_cloudPreferenceKey];
    
    userICloudChoice = [userDefaults boolForKey:_cloudPreferenceKey];
    
    if ([userICloudChoiceSet length] == 0)
        return NO;
    else
        return userICloudChoice;
    
}

- (void)closeDocument
{
    //FLOG(@"closeDocument called.");
    [self saveContext];
    
    [self deregisterForStoreChanges];
    
}

- (void)saveDocument {
    [self saveContext];
}


//////////
/*! Moves a local document to iCloud by migrating the existing store to iCloud and then removing the original store.
 We use a local file name of persistentStore and an iCloud name of persistentStore_ICLOUD so its easy to tell if
 the file is iCloud enabled
 
 */
- (bool)moveStoreToICloud {
    FLOG(@"moveStoreToICloud called");
    
    // Always make a backup of the local store before migrating to iCloud
    [self backupLocalStore];
    
    NSPersistentStoreCoordinator *migrationPSC = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.managedObjectModel];
    
    // Open the store
    id sourceStore = [migrationPSC addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:[self localStoreURL] options:[self localStoreOptions] error:nil];
    
    if (!sourceStore) {
        migrationPSC = nil;
        FLOG(@" failed to add old store");
        return FALSE;
    } else {
        FLOG(@" Successfully added store to migrate");
        
        bool moveSuccess = NO;
        NSError *error;
        
        FLOG(@" About to migrate the store...");
        id migrationSuccess = [migrationPSC migratePersistentStore:sourceStore toURL:[self icloudStoreURL] options:[self icloudStoreOptions] withType:NSSQLiteStoreType error:&error];
        
        if (migrationSuccess) {
            moveSuccess = YES;
            FLOG(@"store successfully migrated");
            [self deregisterForStoreChanges];
            _persistentStoreCoordinator = nil;
            _managedObjectContext = nil;
            sourceStore = nil;
            migrationPSC = nil;
            self.storeURL = [self icloudStoreURL];
            // Now delete the local file
            [self deleteLocalStore];
            return TRUE;
        }
        else {
            sourceStore = nil;
            migrationPSC = nil;
            
            FLOG(@"Failed to migrate store: %@, %@", error, error.userInfo);
            return FALSE;
        }
        
    }
    return FALSE;
}
/*! Moves an iCloud store to local by migrating the iCloud store to a new local store and then removes the store from iCloud.
 
 Note that even if it fails to remove the iCloud files it deletes the local copy.  User may need to clean up orphaned iCloud files using a Mac!
 
 @return Returns YES of file was migrated or NO if not.
 */
- (bool)moveStoreToLocal {
    FLOG(@"moveStoreToLocal called");
    
    // Lets use the existing PSC
    NSPersistentStoreCoordinator *migrationPSC = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.managedObjectModel];
    
    // Open the store
    id sourceStore = [migrationPSC addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:[self icloudStoreURL] options:[self icloudStoreOptions] error:nil];
    
    if (!sourceStore) {
        migrationPSC = nil;
        
        FLOG(@" failed to add old store");
        return FALSE;
    } else {
        FLOG(@" Successfully added store to migrate");
        
        bool moveSuccess = NO;
        NSError *error;
        
        FLOG(@" About to migrate the store...");
        id migrationSuccess = [migrationPSC migratePersistentStore:sourceStore toURL:[self localStoreURL] options:[self localStoreOptions] withType:NSSQLiteStoreType error:&error];
        
        if (migrationSuccess) {
            moveSuccess = YES;
            FLOG(@"store successfully migrated");
            [self deregisterForStoreChanges];
            _persistentStoreCoordinator = nil;
            _managedObjectContext = nil;
            self.storeURL = [self localStoreURL];
            [self removeICloudStore];
            migrationPSC = nil;
            sourceStore = nil;
            return TRUE;
        }
        else {
            migrationPSC = nil;
            sourceStore = nil;
            FLOG(@"Failed to migrate store: %@, %@", error, error.userInfo);
            return FALSE;
        }
        
    }
    
    return TRUE;
}
- (void)removeICloudStore {
    BOOL result;
    NSError *error;
    // Now delete the iCloud content and file
    result = [NSPersistentStoreCoordinator removeUbiquitousContentAndPersistentStoreAtURL:[self icloudStoreURL]
                                                                                  options:[self icloudStoreOptions]
                                                                                    error:&error];
    if (!result) {
        FLOG(@" error removing store");
        FLOG(@" error %@, %@", error, error.userInfo);
        return ;
    } else {
        FLOG(@" Core Data store removed.");
        
        // Now delete the local file
        [self deleteLocalCopyOfiCloudStore];
        
        return ;
    }
    
}
//Check the User setting and if a backup has been requested make one and reset the option
- (bool)backupCurrentStore {
    FLOG(@"backupCurrentStore called");
    
    if (!_makeBackupPreferenceKey) {
        FLOG(@" error _makeBackupPreferenceKey not set!");
        return FALSE;
    }
    
    //[[NSUserDefaults standardUserDefaults] synchronize];
    //NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
    //bool makeBackup = [userDefaults boolForKey:_makeBackupPreferenceKey];
    
    /*
    if (!makeBackup) {
        FLOG(@" backup not required");
        return FALSE;
    }
    */
    
    if ([self userICloudChoice] && [self iCloudStoreExists]) {
        [self backupICloudStore];
    } else if (![self userICloudChoice] && [self localStoreExists]) {
        [self backupLocalStore];
    }
    
    return FALSE;
}
/*! Creates a backup of the ICloud store
 
 @return Returns YES of file was migrated or NO if not.
 */
- (bool)backupICloudStore {
    FLOG(@"backupICloudStore called");
    
    
    // Lets use the existing PSC
    NSPersistentStoreCoordinator *migrationPSC = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.managedObjectModel];
    
    // Open the store
    id sourceStore = [migrationPSC addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:[self icloudStoreURL] options:[self icloudStoreOptions] error:nil];
    
    if (!sourceStore) {
        
        FLOG(@" failed to add old store");
        migrationPSC = nil;
        return FALSE;
    } else {
        FLOG(@" Successfully added store to migrate");
        
        NSError *error;
        
        FLOG(@" About to migrate the store...");
        id migrationSuccess = [migrationPSC migratePersistentStore:sourceStore toURL:[self backupStoreURL] options:[self localStoreOptions] withType:NSSQLiteStoreType error:&error];
        
        if (migrationSuccess) {
            FLOG(@"store successfully backed up");
            migrationPSC = nil;
            // Now reset the backup preference
            [[NSUserDefaults standardUserDefaults] setBool:NO forKey:_makeBackupPreferenceKey];
            [[NSUserDefaults standardUserDefaults] synchronize];
            return TRUE;
        }
        else {
            FLOG(@"Failed to backup store: %@, %@", error, error.userInfo);
            migrationPSC = nil;
            return FALSE;
        }
        
    }
    migrationPSC = nil;
    return FALSE;
}
/*! Creates a backup of the ICloud store
 
 @return Returns YES of file was migrated or NO if not.
 */
- (bool)backupLocalStore {
    FLOG(@"backupLocalStore called");
    
    
    // Lets use the existing PSC
    NSPersistentStoreCoordinator *migrationPSC = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.managedObjectModel];
    
    // Open the store
    id sourceStore = [migrationPSC addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:[self localStoreURL] options:[self localStoreOptions] error:nil];
    
    if (!sourceStore) {
        
        FLOG(@" failed to add old store");
        migrationPSC = nil;
        return FALSE;
    } else {
        FLOG(@" Successfully added store to migrate");
        
        NSError *error;
        
        FLOG(@" About to migrate the store...");
        id migrationSuccess = [migrationPSC migratePersistentStore:sourceStore toURL:[self backupStoreURL] options:[self localStoreOptions] withType:NSSQLiteStoreType error:&error];
        
        if (migrationSuccess) {
            FLOG(@"store successfully backed up");
            migrationPSC = nil;
            // Now reset the backup preference
            [[NSUserDefaults standardUserDefaults] setBool:NO forKey:_makeBackupPreferenceKey];
            [[NSUserDefaults standardUserDefaults] synchronize];
            return TRUE;
        }
        else {
            FLOG(@"Failed to backup store: %@, %@", error, error.userInfo);
            migrationPSC = nil;
            return FALSE;
        }
        
    }
    migrationPSC = nil;
    return FALSE;
}
// In this case we are deleting the local copy of an iCloud document in response
// to detecting that the iCloud file has been removed. We need to do
// the following:
// 1.  Delete the local /CoreDataUbiquitySupport directory using a FileCoordinator or it won't always be
//     removed properly!
//
- (void)deleteLocalCopyOfiCloudStore {
    // We need to get the URL to the store
    FLOG(@"called ");
    
    //FLOG(@" Just returning");
    //return;
    
    NSURL *coreDataSupportFiles = [self localUbiquitySupportURL];
    
    //Check is this is removing the file we are currently migrating
    FLOG(@" Deleting file %@", coreDataSupportFiles);
    
    // Check if the CoreDataUbiquitySupport files exist
    if (![[NSFileManager defaultManager] fileExistsAtPath:[coreDataSupportFiles path]]) {
        FLOG(@" CoreDataUbiquitySupport files do not exist");
        coreDataSupportFiles = nil;
        return;
    }
    
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
        NSFileCoordinator* fileCoordinator = [[NSFileCoordinator alloc] initWithFilePresenter:nil];
        
        [fileCoordinator coordinateWritingItemAtURL:coreDataSupportFiles options:NSFileCoordinatorWritingForDeleting
                                              error:nil byAccessor:^(NSURL* writingURL) {
                                                  NSFileManager* fileManager = [[NSFileManager alloc] init];
                                                  NSError *er;
                                                  //FLOG(@" deleting %@", writingURL);
                                                  bool res = [fileManager removeItemAtURL:writingURL error:&er];
                                                  if (res) {
                                                      FLOG(@"   CoreDataSupport files removed");
                                                      
                                                  }
                                                  else {
                                                      FLOG(@"   CoreDataSupport files  NOT removed");
                                                      FLOG(@"   error %@, %@", er, er.userInfo);
                                                  }
                                              }];
        
    });
    
    return;
}
// In this case we are deleting the local copy of an iCloud document in response
// to detecting that the iCloud file has been removed. We need to do
// the following:
// 1.  Delete the local /CoreDataUbiquitySupport directory using a FileCoordinator or it won't always be
//     removed properly!
//
- (void)deleteLocalStore {
    // We need to get the URL to the store
    FLOG(@"deleteLocalStore called ");
    
    NSURL *fileURL = [self localStoreURL];
    
    //Check is this is removing the file we are currently migrating
    FLOG(@" Deleting file %@", fileURL);
    
    // Check if the CoreDataUbiquitySupport files exist
    if (![[NSFileManager defaultManager] fileExistsAtPath:[fileURL path]]) {
        FLOG(@" Local store file does not exist");
        fileURL = nil;
        return;
    }
    
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
        NSFileCoordinator* fileCoordinator = [[NSFileCoordinator alloc] initWithFilePresenter:nil];
        
        [fileCoordinator coordinateWritingItemAtURL:fileURL options:NSFileCoordinatorWritingForDeleting
                                              error:nil byAccessor:^(NSURL* writingURL) {
                                                  NSFileManager* fileManager = [[NSFileManager alloc] init];
                                                  NSError *er;
                                                  //FLOG(@" deleting %@", writingURL);
                                                  bool res = [fileManager removeItemAtURL:writingURL error:&er];
                                                  if (res) {
                                                      FLOG(@"   Local store file removed");
                                                      
                                                  }
                                                  else {
                                                      FLOG(@"   Local store file  NOT removed");
                                                      FLOG(@"   error %@, %@", er, er.userInfo);
                                                  }
                                              }];
        
    });
    
    return;
}

- (void)saveContext
{
    if (_isOpening) {
        FLOG(@"  application is still busy opening!");
        return;
    }
    
    NSError *error = nil;
    NSManagedObjectContext *managedObjectContext = self.managedObjectContext;
    if (managedObjectContext != nil) {
        if ([managedObjectContext hasChanges] && ![managedObjectContext save:&error]) {
            // Replace this implementation with code to handle the error appropriately.
            // abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
            FLOG(@"Unresolved error %@, %@", error, [error userInfo]);
            
            // Show user message !
            
        }
    }
}
// Returns the managed object context for the application.
// If the context doesn't already exist, it is created and bound to the persistent store coordinator for the application.
- (void)createManagedObjectContext
{
    if (_managedObjectContext != nil) {
        return ;
    }
    
    NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
    if (coordinator != nil) {
        _managedObjectContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSMainQueueConcurrencyType];
        [_managedObjectContext setPersistentStoreCoordinator:coordinator];
    }
    
}

// Returns the managed object context for the application (which is already bound to the persistent store coordinator for the application.)
- (NSManagedObjectContext *)managedObjectContext
{
    if (_managedObjectContext) {
        return _managedObjectContext;
    }
    
    NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
    if (!coordinator) {
        NSMutableDictionary *dict = [NSMutableDictionary dictionary];
        [dict setValue:@"Failed to initialize the store" forKey:NSLocalizedDescriptionKey];
        [dict setValue:@"There was an error building up the data file." forKey:NSLocalizedFailureReasonErrorKey];
        NSError *error = [NSError errorWithDomain:@"YOUR_ERROR_DOMAIN" code:9999 userInfo:dict];
        //[[NSApplication sharedApplication] presentError:error];
        FLOG(@" error failed to initialise the store, %@", error);
        return nil;
    }
    _managedObjectContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSMainQueueConcurrencyType];
    [_managedObjectContext setPersistentStoreCoordinator:coordinator];
    
    return _managedObjectContext;
}

// Returns the managed object model for the application.
// If the model doesn't already exist, it is created from the application's model.
- (NSManagedObjectModel *)managedObjectModel
{
    if (_managedObjectModel != nil && !_use_old_model) {
        return _managedObjectModel;
    }
    NSURL *modelURL;
    if (_use_old_model) {
        if (!_sourceModel) {
            FLOG(@" ERROR NO SOURCE MODEL FOR MIGRATION !");
            exit(0);
        }
        //modelURL = [[NSBundle mainBundle] URLForResource:@"CoreDataLibraryApp_old" withExtension:@"momd"];
        _managedObjectModel = _sourceModel;
    }
    else {
        modelURL = [[NSBundle mainBundle] URLForResource:@"CoreDataLibraryApp" withExtension:@"momd"];
    
        _managedObjectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];
    }
    return _managedObjectModel;
}
- (bool)storeExists {
    FLOG(@" called");
    
    if (_isCloudEnabled) {
        
        return [self iCloudStoreExists];
        
    } else {
        
        FLOG(@"  storeURL is %@", [[self storeURL] path]);
        BOOL isDir;
        BOOL fileExists = [[NSFileManager defaultManager] fileExistsAtPath:[[self storeURL] path] isDirectory:&isDir];
        
        FLOG(@"  storeURL %@", (fileExists ? @"exists" : @"does not exist"));
        
        return fileExists;
    }
}
- (void)openPersistentStore {
    FLOG(@" called... XXXXXXXXXX XXX XXX XX XX X X");
    _isOpening = YES;
    
    NSManagedObjectModel *mom = [self managedObjectModel];
    if (!mom) {
        FLOG(@" No model to generate a store from");
        _isOpening = NO;
        return ;
    }
    
    // Check if the store exists, if not then set _load_seed_data flag because this must be a first time
    if (!_isUpgrading && ![self storeExists]) {
        FLOG(@" setting _load_seed_data = YES");
        _load_seed_data = YES;
    }

    
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSURL *applicationFilesDirectory = [self applicationFilesDirectory];
    NSError *error = nil;
    
    NSDictionary *properties = [applicationFilesDirectory resourceValuesForKeys:@[NSURLIsDirectoryKey] error:&error];
    
    if (!properties) {
        BOOL ok = NO;
        if ([error code] == NSFileReadNoSuchFileError) {
            ok = [fileManager createDirectoryAtPath:[applicationFilesDirectory path] withIntermediateDirectories:YES attributes:nil error:&error];
        }
        if (!ok) {
            FLOG(@" error creating directory");
            dispatch_async(dispatch_get_main_queue(), ^{
                [[NSApplication sharedApplication] presentError:error];
            });
            return;
        }
    } else {
        if (![properties[NSURLIsDirectoryKey] boolValue]) {
            // Customize and localize this error.
            NSString *failureDescription = [NSString stringWithFormat:@"Expected a folder to store application data, found a file (%@).", [applicationFilesDirectory path]];
            
            NSMutableDictionary *dict = [NSMutableDictionary dictionary];
            [dict setValue:failureDescription forKey:NSLocalizedDescriptionKey];
            error = [NSError errorWithDomain:@"YOUR_ERROR_DOMAIN" code:101 userInfo:dict];
            FLOG(@" error file found expecting directory");
            dispatch_async(dispatch_get_main_queue(), ^{
                [[NSApplication sharedApplication] presentError:error];
            });
            return;
        }
    }
    FLOG(@" creating persistentStoreCoordinator");
    NSPersistentStoreCoordinator *coordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:mom];
    [self registerForStoreChanges:coordinator];
    
    FLOG(@" adding persistentStore to Coordinator");
    FLOG(@" storeURL is %@", self.storeURL);
    FLOG(@" storeOptions are %@", self.storeOptions);
    
    if (![coordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:self.storeURL options:self.storeOptions error:&error]) {
        FLOG(@" error adding PersistentStore %@", error);
        dispatch_async(dispatch_get_main_queue(), ^{
            [[NSApplication sharedApplication] presentError:error];
        });
        return;
    }
    _persistentStoreCoordinator = coordinator;
    
    NSURL *url = [[[_persistentStoreCoordinator persistentStores] objectAtIndex:0] valueForKey:@"URL"];
    
    FLOG(@" STORE FILE is %@", [url path]);
    [[NSUserDefaults standardUserDefaults] setURL:url forKey:@"iCloud Store File Name"];
    
    // Now load any seed data
    [self loadSeedDataIfRequired];
    
    // Tell the world the store has opened
    FLOG(@" post notifications");
    [self postStoreReadyNotification];
    //[self postStoreChangedNotification];
    //[self createTimer];
    _isOpening = NO;
    FLOG(@" FINISHED... XXXXXXXXXX XXX XXX XX XX X X");
    
}

// Returns the persistent store coordinator for the application.
// If the coordinator doesn't already exist, it is created and the application's store added to it.
- (NSPersistentStoreCoordinator *)persistentStoreCoordinator
{
    FLOG(@"persistentStoreCoordinator called");
    
    if (_persistentStoreCoordinator != nil) {
        return _persistentStoreCoordinator;
    }
    
    return _persistentStoreCoordinator;
    
}

- (void)snooze:(int)seconds
{
    FLOG(@"  sleeping for %d seconds... ", seconds);
    sleep(seconds);
    FLOG(@"  done sleeping.");
}

// Returns the local store options
- (NSDictionary*)localStoreOptions {
    return @{NSMigratePersistentStoresAutomaticallyOption:@YES,
             NSInferMappingModelAutomaticallyOption:@YES,
             NSSQLitePragmasOption:@{ @"journal_mode" : @"DELETE" }};
}
// Returns the iCloud store options
- (NSDictionary*)icloudStoreOptions {
    
    NSString *icloudFilename = [NSString stringWithFormat:@"%@_ICLOUD", [OSxCDStackManager persistentStoreName]];
    
    return @{NSPersistentStoreUbiquitousContentNameKey:icloudFilename,
             NSMigratePersistentStoresAutomaticallyOption:@NO,
             NSInferMappingModelAutomaticallyOption:@NO,
             NSSQLitePragmasOption:@{ @"journal_mode" : @"DELETE" }};
}
// Checks if the file is iCloud or Local and returns the required options
- (NSDictionary*)storeOptions {
    
    NSString *string = [[self.storeURL URLByDeletingPathExtension] lastPathComponent];
    
    // If it's an iCloud file
    if ([string rangeOfString:@"_ICLOUD"].location != NSNotFound) {
        return [self icloudStoreOptions];
        
    } else {
        return [self localStoreOptions];
        
    }
    
}
- (NSArray *)getData:(NSString*)entityName sortField:(NSString*)sortKey predicate:(NSPredicate*)predicate
{
    return [self getData:entityName sortField:sortKey predicate:predicate managedObjectContext:[self managedObjectContext]];
}

- (NSArray *)getData:(NSString*)entityName sortField:(NSString*)sortKey predicate:(NSPredicate*)predicate managedObjectContext:(NSManagedObjectContext*)context
{
    FLOG(@"getData called");
    
    if (context == nil) {
        FLOG(@"Error can't continue with null managedObjectContext");
        return nil;
    }
    if (entityName == nil) {
        FLOG(@"Error can't continue with null entityName");
        return nil;
    }
    if (sortKey == nil) {  // if its not set then just set it as follows...
        FLOG(@"Error can't continue with null sortField");
        return nil;
    }
    FLOG(@" entity is %@", entityName);
    
    NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
    // Edit the entity name as appropriate.
    NSEntityDescription *entity = [NSEntityDescription entityForName:entityName inManagedObjectContext:context];
    if (entity == nil) {
        FLOG(@"  error finding entity %@ in class %@", entityName, [self class]);
        return nil;
    }
    
    [fetchRequest setEntity:entity];
    
    if (predicate)
        [fetchRequest setPredicate:predicate];
    
    // Edit the sort key as appropriate.
    NSSortDescriptor *sortDescriptor = [[NSSortDescriptor alloc] initWithKey:sortKey ascending:YES];
    NSArray *sortDescriptors =[NSArray arrayWithObjects:sortDescriptor,nil];
    [fetchRequest setSortDescriptors:sortDescriptors];
    
    NSError *error = nil;
    NSArray *result = [context executeFetchRequest:fetchRequest error:&error];
    
	if (!result) {
        // Replace this implementation with code to handle the error appropriately.
        // abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
	    NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
	    //abort();
	}
    
    return result;
}

/*! Checks to see whether the user has previously selected the iCloud storage option, and if so then check
 whether the iCloud identity has changed (i.e. different iCloud account being used or logged out of iCloud).
 
 If the user has previously chosen to use iCloud and we're still signed in, setup the CloudManager
 with cloud storage enabled.
 
 If iCloud is available AND if no user choice is recorded, use a UIAlert to fetch the user's preference.
 
 If iCloud is available AND if user has selected to Use iCloud then check if any local files need to be
 migrated.
 
 if iCloud is available AND if user has selected to NO Use iCloud then check if any iCloud files need to
 be migrated to local storage.
 
 */
- (void)checkUserICloudPreferenceAndSetupIfNecessary
{
    dispatch_async(dispatch_get_main_queue(), ^{
        [self performChecks];
    });
}
/**  On startup check if a local file exists and open it, if not
 */
-(void)performChecks {
    FLOG(@"checkUserICloudPreferenceAndSetupIfNecessary called");
    
    // Check if a backup has been requested
    //[self backupCurrentStore];
    
    [self setUbiquityID:_ubiquityContainerKey ];
    
    id currentToken = [[NSFileManager defaultManager] ubiquityIdentityToken];
    FLOG(@" ubiquityIdentityToken is %@", currentToken);
    FLOG(@" iCloud container is %@", [self iCloudCoreDataURL]);
    
    if (!currentToken) {
        FLOG(@" iCloud is not enabled");
        // If there is no token now, set our state to NO
        //[self setIsCloudEnabled:NO];
    } else {
        FLOG(@" iCloud is enabled");
        if (self.isCloudEnabled) {
            [[NSNotificationCenter defaultCenter] postNotificationName:ICloudStateUpdatedNotification object:nil];
        }
    }
    
    [[NSUserDefaults standardUserDefaults] synchronize];
    NSString* userICloudChoiceSet = [[NSUserDefaults standardUserDefaults] stringForKey:_cloudPreferenceSet];
    
    // Call this twice because sometimes the first call returns incorrect value!
    bool userICloudChoice = [[NSUserDefaults standardUserDefaults] boolForKey:_cloudPreferenceKey];
    
    if (!([userICloudChoiceSet length] == 0)) {
        FLOG(@" User preference for %@ is %@", _cloudPreferenceKey, (userICloudChoice ? @"YES" : @"NO"));
    }
    
    if (userICloudChoice) {
        
        FLOG(@" User selected iCloud");
        _useICloudStorage = YES;
        
        // Display notice if previous iCloud account is not available
        [self checkUbiquitousTokenFromPreviousLaunch:currentToken];
        
        
    } else {
        
        FLOG(@" User disabled iCloud");
        _useICloudStorage = NO;
        
    }
    
    // iCloud is active
    if (currentToken) {
        
        FLOG(@" iCloud is active");
        
        // If user has not yet set preference then prompt for them to select a preference
        if ([userICloudChoiceSet length] == 0) {
            
            FLOG(@" userICloudChoiceSet has not been set yet, so ask the user what they want to do");
            [self setIsCloudEnabled:[self cloudChoiceAlert]];
            
        }
        else {
            FLOG(@" userICloudChoiceSet is set");
            if (userICloudChoice) {
                FLOG(@" userICloudChoice is YES");
                // iCloud is available and user has selected to use it
                // Check if any local files need to be migrated to iCloud
                // and migrate them
                [self setIsCloudEnabled:YES];
                
            } else  {
                FLOG(@" userICloudChoice is NO");
                // iCloud is available but user has chosen to NOT Use iCloud
                // Check that NO local file exists already
                if (![self localStoreExists]) {
                    // and IF an iCloud file exists
                    if ([self iCloudStoreExists]) {
                        FLOG(@" iCloudStoreExists exists");
                        
                        //  Ask the user if they want to migrate the iCloud file to a local file
                        int res = [self promptUserAboutICloudDocumentStorage];
                        if (res == STORE_KEEP) {
                            FLOG(@" 'Keep using iCloud' selected");
                            FLOG(@" turn Use iCloud back ON");
                            [[NSUserDefaults standardUserDefaults] setBool:YES forKey:_cloudPreferenceKey];
                            [[NSUserDefaults standardUserDefaults] synchronize];
                            _useICloudStorage = YES;
                            [self setIsCloudEnabled:YES];
                            
                        }
                        else if (res == STORE_MOVE) {
                            FLOG(@" 'Keep on My Mac' selected");
                            FLOG(@" copy to local storage");
                            _useICloudStorage = NO;
                            [self setDeleteICloudFiles:NO];
                            [self setIsCloudEnabled:NO];
                            
                        }else {
                            FLOG(@" 'Delete from My Mac' selected");
                            FLOG(@" delete copies from iPhone");
                            _useICloudStorage = NO;
                            [self setDeleteICloudFiles:YES];
                            [self setIsCloudEnabled:NO];
                        }

                        
                    } else {
                        // And because no local file exists already this must the the first time therefor we need to load seed data
                        _load_seed_data = YES;
                        
                        // Otherwise just set iCloud enabled = NO
                        [self setIsCloudEnabled:NO];
                        
                    }
                } else {
                    // A Local file already exists so what to do ?
                    // Just tell the user a file has been detected in iCloud
                    // and ask if they want to start using the iCloud file
                    [self setIsCloudEnabled:NO];
                }
            }
        }
    }
    else {
        FLOG(@" iCloud is not active");
        [self setIsCloudEnabled:NO];
        _useICloudStorage = NO;
        [[NSUserDefaults standardUserDefaults] setBool:NO forKey:_cloudPreferenceKey];
        [[NSUserDefaults standardUserDefaults] synchronize];
        
        // Since the user is signed out of iCloud, reset the preference to not use iCloud, so if they sign in again we will prompt them to move data
        [[NSUserDefaults standardUserDefaults] removeObjectForKey:_cloudPreferenceSet];
    }
    [self storeCurrentUbiquityToken:[[NSFileManager defaultManager] ubiquityIdentityToken]];
    
}
- (void)setICloudPreference:(bool)pref {
    [[NSUserDefaults standardUserDefaults] setBool:pref forKey:_cloudPreferenceKey];
    [[NSUserDefaults standardUserDefaults] setValue:@"YES" forKey:_cloudPreferenceSet];
    
    [[NSUserDefaults standardUserDefaults] synchronize];
    [self setIsCloudEnabled:pref];
    
}
/** Prompts user for storage option if its not set
 
 */
- (bool)cloudChoiceAlert {
    NSLog(@"cloudChoiceAlert called");
    
    NSString *question = NSLocalizedString(@"Should documents be stored in iCloud or on just this device?", @"Choose Storage Option question");
    NSString *info = NSLocalizedString(@"Choose a storage option", @"Choose Storage Option info");
    NSString *localButton = NSLocalizedString(@"Local only", @"Local only button title");
    NSString *icloudButton = NSLocalizedString(@"iCloud", @"iCloud button title");
    NSAlert *alert = [[NSAlert alloc] init];
    [alert setMessageText:question];
    [alert setInformativeText:info];
    [alert addButtonWithTitle:localButton];
    [alert addButtonWithTitle:icloudButton];
    
    NSInteger answer = [alert runModal];
    
    if (answer == NSAlertFirstButtonReturn) {
        // Set the preference
        [[NSUserDefaults standardUserDefaults] setBool:NO forKey:_cloudPreferenceKey];
        [[NSUserDefaults standardUserDefaults] setValue:@"NO" forKey:_cloudPreferenceSet];
        _useICloudStorage = NO;
        return NO;
    }
    else {
        // Set the preference
        [[NSUserDefaults standardUserDefaults] setBool:YES forKey:_cloudPreferenceKey];
        [[NSUserDefaults standardUserDefaults] setValue:@"YES" forKey:_cloudPreferenceSet];
        _useICloudStorage = YES;
        return YES;
    }
}
/** Prompts user for storage option if its not set
 
 */
- (bool)mergeWithICloudChoiceAlert {
    FLOG(@" called");
    
    NSString *question = NSLocalizedString(@"iCloud file detected, choose an option.", @"Choose Storage Option question");
    NSString *info = NSLocalizedString(@"Choose a storage option", @"Choose Storage Option info");
    NSString *mergeButton = NSLocalizedString(@"Merge data", @"Merge button title");
    NSString *icloudButton = NSLocalizedString(@"Use the iCloud data", @"iCloud button title");
    NSAlert *alert = [[NSAlert alloc] init];
    [alert setMessageText:question];
    [alert setInformativeText:info];
    [alert addButtonWithTitle:mergeButton];
    [alert addButtonWithTitle:icloudButton];
    
    NSInteger answer = [alert runModal];
    
    if (answer == NSAlertFirstButtonReturn) {
        return YES;
    }
    else {
        return NO;
    }
}
- (void)performApplicationWillEnterForegroundCheck
{
    FLOG(@"applicationWillEnterForegroundCheck called");
    
    if (_useICloudStorage == -1) {
        
        FLOG(@"_useICloudStorage is %@", (_useICloudStorage ? @"YES" : @"NO"));
    }else {
        FLOG(@" _useICloudStorage is NIL! XXXXXXXX");
    }
    
    if (_isOpening) {
        FLOG(@" store is busy opening!");
        return;
    }
    
    // Check if a backup has been requested
    [self backupCurrentStore];
    
    // Check if the app settings have been changed in the Settings Bundle (we use a Settings Bundle which
    // shows settings in the Devices Settings app, along with all the other device settings).
    [[NSUserDefaults standardUserDefaults] synchronize];
    NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
    bool userICloudChoice = [userDefaults boolForKey:_cloudPreferenceKey];
    
    if (_useICloudStorage != userICloudChoice) {
        FLOG(@"  Use iCloud preference has changed to %@", (userICloudChoice ? @"YES" : @"NO"));
    }else {
        FLOG(@" Use iCloud preference has NOT changed");
        return;
    }
    
    // Check if iCloud is available, if its not then we can't do anything anyway
    
    if ([self isICloudAvailable]) {
        
        // iCloud option has been turned off
        if (!userICloudChoice) {
            
            // If iCloud files still exist then ask what to do
            if ([self iCloudStoreExists]) {
                
                [self promptUserAboutICloudDocumentStorage];
                return;
                
            } else {
                [self setIsCloudEnabled:NO];
            }
            // Handle the users response in the alert callback
            
        } else {
            
            // iCloud is turned on so just copy them across... including the one we may have open
            
            //LOG(@" iCloud turned on so copy any created files across");
            [self setIsCloudEnabled:YES];  // This does all the work for us
            
            _useICloudStorage = YES;
            
        }
    }
    
}
/*! This method is called when Use iCloud preference has been turned OFF to ask the user whether they want to keep iCloud files, delete them or keep using iCloud
 
 */
- (int)promptUserAboutICloudDocumentStorage {
        FLOG(@" called");
        
        NSString *question = NSLocalizedString(@"You're not using iCloud but no local file exists and an iCloud file has been detected, what would your like to do? ", @"Choose Storage Option question");
        NSString *info = NSLocalizedString(@"Choose a storage option", @"Choose Storage Option info");
        NSString *icloudButton = NSLocalizedString(@"Keep using iCloud", @"iCloud button title");
        NSString *keepButton = NSLocalizedString(@"Move to my Mac", @"Keep button title");
        NSString *deleteButton = NSLocalizedString(@"Delete from my Mac", @"Keep button title");
        NSAlert *alert = [[NSAlert alloc] init];
        [alert setMessageText:question];
        [alert setInformativeText:info];
        [alert addButtonWithTitle:icloudButton];
        [alert addButtonWithTitle:keepButton];
        [alert addButtonWithTitle:deleteButton];

        NSInteger answer = [alert runModal];
        
        if (answer == NSAlertFirstButtonReturn) {
            return STORE_KEEP;
        }
        else if (answer == NSAlertSecondButtonReturn) {
            return STORE_MOVE;
        }else {
            return STORE_DELETE;
        }
}
- (void)loadDataInBackground {
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
        [self loadData];
    });
    
}
/** This uses performBlock but ends up blocking the main thread!
 Duh - what were they thinking!
 */
- (void)loadData2 {
    FLOG(@"loadData called");
    FLOG(@" waiting 5 seconds...");
    sleep(5);
    [self showBackgroundTaskActive];
    
    FLOG(@" starting load...");
    NSManagedObjectContext *bgContext = self.managedObjectContext;
    
    [bgContext performBlockAndWait:^(void){
        for (int i = 1; i<=2; i++) {
            // Add a few companies
            [self insertNewCompany:bgContext count:2];
            [bgContext processPendingChanges];
            
            // Save the context.
            
            NSError *error = nil;
            if (![bgContext save:&error]) {
                FLOG(@"  Unresolved error %@, %@", error, [error userInfo]);
            }
            
            FLOG(@"   waiting 10 seconds...");
            sleep(10);
        }
    }];
    FLOG(@" loading ended...");
    //[self showBackgroundTaskInactive];
    
}
/** This load on the background thread and posts notifications when done
 */
- (void)loadData {
    FLOG(@"loadData called");
    FLOG(@" waiting 5 seconds...");
    sleep(5);
    [self showBackgroundTaskActive];
    
    NSManagedObjectContext *bgContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSConfinementConcurrencyType];
    // Register for saves in order to merge any data from background threads
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesDidSave:) name: NSManagedObjectContextDidSaveNotification object:bgContext];

    while (self.persistentStoreCoordinator == nil) {
        FLOG(@" persistentStoreCoordinator = nil, waiting 5 seconds to try again...");
        sleep(5);
    }
    
    bgContext.persistentStoreCoordinator = [self persistentStoreCoordinator];
    
    FLOG(@" starting load...");
    
    _companyCount = [self getCompanyCount:bgContext] + 1;
    _personCount = [self getCompanyCount:bgContext] + 1;

    
    for (int i = 1; i<=1; i++) {
        // Add a few companies
        [self insertNewCompany:bgContext count:100];
        [bgContext processPendingChanges];
        
        // Save the context.
        NSError *error = nil;
        if (![bgContext save:&error]) {
            FLOG(@"  Unresolved error %@, %@", error, [error userInfo]);
        }
        //FLOG(@"   waiting 2 seconds...");
        sleep(0.01);
    }
    // Register for saves in order to merge any data from background threads
    [[NSNotificationCenter defaultCenter] removeObserver:self name: NSManagedObjectContextDidSaveNotification object:bgContext];

    FLOG(@" loading ended...");
    [self showBackgroundTaskInactive];
    [self postJobDoneNotification];

}
-(long)getCompanyCount:(NSManagedObjectContext*)moc {
    return [[self getData:@"Company" sortField:@"name" predicate:nil managedObjectContext:moc] count];
}
-(long)getPeopleCount:(NSManagedObjectContext*)moc {
    return [[self getData:@"Person" sortField:@"name" predicate:nil managedObjectContext:moc] count];
}

- (void)insertNewCompany:(NSManagedObjectContext *)moc count:(int)count
{
    //FLOG(@"insertNewCompany:count: called");
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    [dateFormatter setDateFormat:@"yyyyMMddHHmmssSSS"];
    NSManagedObjectContext *context = moc;
    
    for (int i = 1; i <= count; i++) {
        NSManagedObject *newManagedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Company" inManagedObjectContext:context];
        
        NSString *name = [NSString stringWithFormat:@"New Company %06ld", _companyCount++];
        
        [newManagedObject setValue:name forKey:@"name"];
        //FLOG(@"  created company %@", name);
        [self insertNewEmployees:context company:newManagedObject count:10];
    }
}
- (void)insertNewEmployees:(NSManagedObjectContext *)moc company:(NSManagedObject*)company count:(int)count
{
    //FLOG(@"insertNewEmployees:count: called");
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    [dateFormatter setDateFormat:@"yyyyMMddHHmmssSSS"];
    NSManagedObjectContext *context = moc;
    
    for (int i = 1; i <= count; i++) {
        NSManagedObject *newManagedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Person" inManagedObjectContext:context];
        
        NSString *name = [NSString stringWithFormat:@"New Person %06ld", _personCount++];
        
        [newManagedObject setValue:name forKey:@"name"];
        [newManagedObject setValue:company forKey:@"employer"];
        //FLOG(@"  created person %@", name);
    }
}
- (void)insertNewProjectType:(NSManagedObjectContext *)moc name:(NSString*)projectType
{
    FLOG(@"insertNewProjectType:count: called");
    NSManagedObjectContext *context = moc;
    
    NSManagedObject *newManagedObject = [NSEntityDescription insertNewObjectForEntityForName:@"ProjectType" inManagedObjectContext:context];
    
    [newManagedObject setValue:projectType forKey:@"name"];
    FLOG(@"  created project type %@", projectType);
}
- (void)loadSeedDataIfRequired {
    FLOG(@" called");
    if (_load_seed_data) {
      [self loadSeedData];
        _load_seed_data = NO;
    }
}
- (void)loadSeedData {
    FLOG(@" called");
    
    NSManagedObjectContext *bgContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSConfinementConcurrencyType];
    // Register for saves in order to merge any data from background threads
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesDidSave:) name: NSManagedObjectContextDidSaveNotification object:bgContext];
    
    while (self.persistentStoreCoordinator == nil) {
        FLOG(@" persistentStoreCoordinator = nil, waiting 5 seconds to try again...");
        sleep(5);
    }
    
    bgContext.persistentStoreCoordinator = [self persistentStoreCoordinator];

    
    [self insertNewProjectType:bgContext name:@"Application"];
    [self insertNewProjectType:bgContext name:@"Infrastructure"];
    [self insertNewProjectType:bgContext name:@"Process"];
    [self insertNewProjectType:bgContext name:@"Products & Services"];
    
    [bgContext processPendingChanges];
    NSError *error = nil;
    if (![bgContext save:&error]) {
        FLOG(@"  Unresolved error %@, %@", error, [error userInfo]);
    }

}
- (void)showBackgroundTaskActive {
}
- (void)showBackgroundTaskInactive {
    
}

// Returns the directory the application uses to store the Core Data store file. This code uses a directory named "au.com.ossh.CoreDataLibraryAppOSX" in the user's Application Support directory.
- (NSURL *)applicationFilesDirectory
{
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSURL *appSupportURL = [[fileManager URLsForDirectory:NSApplicationSupportDirectory inDomains:NSUserDomainMask] lastObject];
    return [appSupportURL URLByAppendingPathComponent:@"au.com.ossh.CoreDataLibraryAppOSX"];
}
- (void)deleteDataInBackground {
    
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
        [self deleteData];
    });
    
}
- (void)deleteData {
    FLOG(@"deleteData called");
    FLOG(@" waiting 5 seconds...");
    sleep(5);
    [self showBackgroundTaskActive];
    
    NSManagedObjectContext *bgContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
    
    // Register for saves in order to merge any data from background threads
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(storesDidSave:) name: NSManagedObjectContextDidSaveNotification object:bgContext];
    

    while (self.persistentStoreCoordinator == nil) {
        FLOG(@" persistentStoreCoordinator = nil, waiting 5 seconds to try again...");
        sleep(5);
    }
    
    bgContext.persistentStoreCoordinator = [self persistentStoreCoordinator];
    
    FLOG(@" fetching data...");
    
    NSArray *companies = [self getData:@"Company" sortField:@"name" predicate:nil managedObjectContext:bgContext];
    
    NSUInteger count = companies.count;
    
    if (count>2) {
        for (int i = 0; i<3; i++) {
            NSManagedObject *object = [companies objectAtIndex:i];
            FLOG(@" deleting %@", [object valueForKey:@"name"]);
            [bgContext deleteObject:object];
            [bgContext processPendingChanges];
            NSError *error = nil;
            if (![bgContext save:&error]) {
                FLOG(@"  Unresolved error %@, %@", error, [error userInfo]);
            }
            FLOG(@"   waiting 5 seconds...");
            sleep(5);
        }
        
    }
    
    [[NSNotificationCenter defaultCenter] removeObserver:self name: NSManagedObjectContextDidSaveNotification object:bgContext];
    
    /*
    dispatch_async(dispatch_get_main_queue(),^(void){
        [[NSNotificationCenter defaultCenter] removeObserver:self name: NSManagedObjectContextDidSaveNotification object:nil];
    });
    */
    
    FLOG(@" delete ended...");
    [self showBackgroundTaskInactive];
    [self postJobDoneNotification];
}
- (void)saveFile:(NSURL *)fileURL {
    FLOG(@" called with %@", fileURL);
    [self importFile:fileURL];
}
/**  Copies iCloud file to Local container, fails if target file exists
 
 @param fileURL  URL of the file to be copied
 */
- (void)importFile:(NSURL*)fileURL {
    FLOG(@" called with %@", fileURL);
   // @autoreleasepool {
        NSFileCoordinator *fc = [[NSFileCoordinator alloc] initWithFilePresenter:nil];
        
        NSURL *sourceURL = fileURL;
        NSString *filename = [fileURL lastPathComponent];
    FLOG(@" source file is %@", sourceURL);
    
    NSURL *appDirectory = [self applicationFilesDirectory];
    FLOG(@" appDirectory is %@", appDirectory);
    
        __block NSURL *destinationURL = [[self applicationFilesDirectory] URLByAppendingPathComponent:filename];
        
        
        FLOG(@" target file is %@", destinationURL);
        
        NSError *cError;
        
        [fc coordinateReadingItemAtURL:sourceURL options:NSFileCoordinatorReadingWithoutChanges error:&cError byAccessor:^(NSURL *sourceURLToUse) {
            NSError *error = nil;
            NSFileManager *fm = [[NSFileManager alloc] init];
            
            if ([fm fileExistsAtPath:[destinationURL path]]) {
                FLOG(@" target file exists");
                NSURL * newURL = [self getNewFileURL:destinationURL local:YES];
                destinationURL = newURL;
                FLOG(@" new target file is %@", destinationURL);
            }
            
            //simply copy the file over
            BOOL copySuccess = [fm copyItemAtPath:[sourceURLToUse path]
                                           toPath:[destinationURL path]
                                            error:&error];
            if (copySuccess) {
                FLOG(@" copied file successfully");
                //[self postFileUpdateNotification];
            } else {
                FLOG(@"Error copying items Error: %@, %@", error, error.userInfo);
            }
        }];
        
        if (cError) {
            FLOG(@" error is %@", cError);
        }
        
        fc = nil;
    //}
}
- (BOOL)docNameExistsInLocalURLs:(NSURL *)docName {
    FLOG(@" called");
    //BOOL nameExists = NO;
    NSFileManager *fm = [[NSFileManager alloc] init];
    return [fm fileExistsAtPath:[docName path]];
}
- (BOOL)docNameExistsInICloudURLs:(NSString *)docName {
    BOOL nameExists = NO;
    for (NSURL * fileURL in _iCloudURLs) {
        if ([[fileURL lastPathComponent] isEqualToString:docName]) {
            nameExists = YES;
            break;
        }
    }
    return nameExists;
}
// Check if filename exists locally or in iCloud
- (NSURL*)getNewFileURL:(NSURL *)fileURL local:(BOOL)local {
    FLOG(@" called");
    NSInteger docCount = 0;
    NSString* newDocName = nil;
    NSString * extension = [fileURL pathExtension];
    NSString * prefix = [[fileURL lastPathComponent] stringByDeletingPathExtension];
    
    // At this point, the document list should be up-to-date.
    BOOL done = NO;
    BOOL first = YES;
    while (!done) {
        if (first) {
            first = NO;
            newDocName = [NSString stringWithFormat:@"%@.%@",
                          prefix, extension];
        } else {
            newDocName = [NSString stringWithFormat:@"%@ %ld.%@",
                          prefix, docCount, extension];
        }
        
        // Look for an existing document with the same name. If one is
        // found, increment the docCount value and try again.
        BOOL nameExists;
        if (local) {
            nameExists = [self docNameExistsInLocalURLs:[[fileURL URLByDeletingLastPathComponent] URLByAppendingPathComponent:newDocName]];
        } else {
            nameExists = [self docNameExistsInICloudURLs:newDocName];
        }
        if (!nameExists) {
            break;
        } else {
            docCount++;
        }
        
    }
    
    NSURL *newURL = [[fileURL URLByDeletingLastPathComponent] URLByAppendingPathComponent:newDocName];
    
    return newURL;
}
/** Check the Core Data version to see if its compatible or if a lightweight migration is required
    If a lightweight migration is required then migrate the iCloud store and open it then migrate it back to iCloud
 
    Return Returns YES if model needs to be upgraded and NO if not
 */
- (bool)checkCDModelVersion {
    FLOG(@" called");
    
    // We don't need to do this if iCloud is not enabled
    if (!self.isCloudEnabled) {
        FLOG(@" iCloud is not enabled");
        return YES;
    }

    // First find the store file - and there seems to be no easy way to determine what this filename is.
    // The only reliable way is to open it with the old version of the model (to ensure Core Data does not
    // start the upgrade !  The other way is to try and figure out how Core Data builds the path.
    
    
    NSManagedObjectModel *model = [self managedObjectModel];
    
    //FLOG(@" app model is %@", model);
    //FLOG(@" app model entity version hashes are %@", [model entityVersionHashesByName]);
    
    NSURL * storeURL = [[NSUserDefaults standardUserDefaults] URLForKey:@"iCloud Store File Name"];
    
    FLOG(@" storeURL is %@", [storeURL path]);
    // File must exist
    if ([[NSFileManager defaultManager] fileExistsAtPath:[storeURL path]]) {
        FLOG(@" file exists :-)");
    }
    else {
        FLOG(@" file does not exist :-(");
        return NO;
    }
    
    if (storeURL == nil)
        return NO;
    
    NSError *error;
    NSDictionary *metaData = [NSPersistentStoreCoordinator metadataForPersistentStoreOfType:NSSQLiteStoreType URL:storeURL error:&error];
    
    if (!metaData) {
        FLOG(@" problem getting metaData");
        FLOG(@"  - error is %@, %@", error, error.userInfo);
        return NO;
    }
    
    bool result = [model isConfiguration:nil compatibleWithStoreMetadata:metaData];
    if (!result) {
        FLOG(@" file is not compatible!");
        FLOG(@" metadata is %@", metaData);
        _sourceModel =
        [NSManagedObjectModel mergedModelFromBundles:nil
                                    forStoreMetadata:metaData];
        FLOG(@" source model is %@", _sourceModel);

        return YES;
    } else {
        FLOG(@" file is compatible!");
        
    }
    
    return !result;
    
}
@end
